module set_all_extr_m

  implicit none

contains

  subroutine set_all_extr(extr_map, list, periodic, ssh)

    ! This procedure finds all the extrema of ssh.

    use config_m, only: max_radius
    use derived_types, only: eddy
    use input_ssh_m, only: nlon, nlat
    use invert_proj_m, only: invert_proj
    use local_extrema_m, only: local_extrema

    integer, allocatable, intent(out):: extr_map(:, :)
    ! (1 - max_radius(1):nlon + max_radius(1), nlat) if the grid is
    ! periodic in longitude, else (nlon, nlat). At a point of extremum
    ! SSH: identification number or this extremum. 0 at other points.

    type(eddy), allocatable, intent(out):: list(:) ! (number_extr)
    ! "Eddies" at a given date. In general, there may be both cyclonic
    ! and anticyclonic eddies in the list. Eddies include "eddies"
    ! without an outermost contour, that is, only the extremum is
    ! defined. The subscript value in list is the "identification
    ! number" of the extremum. Define only list%extr, list%cyclone,
    ! list%innermost_level. It would be difficult to replace argument
    ! "list" by arguments corresponding to the defined components of
    ! "list" because of allocations.

    logical, intent(in):: periodic ! grid is periodic in longitude

    real, intent(in):: ssh(1 - merge(max_radius(1), 0, periodic):, :)
    ! (1 - max_radius(1):nlon + max_radius(1), nlat) if the grid is periodic
    ! in longitude, else (nlon, nlat). Sea-surface height, in m.

    ! Local:

    logical, allocatable:: cyclone(:) ! (number_extr)
    integer i, copy, number_extr

    integer, allocatable:: ind_extr(:, :) ! (2, n_extr)
    ! List of couple of subscripts of extrema of ssh, in the two
    ! dimensions.

    real, allocatable:: innermost_level(:) ! (number_extr)
    ! SSH level of the innermost contour around each extremum, in
    ! m. By construction, innermost_level < extremum for a maximum, >
    ! extremum for a minimum.

    !--------------------------------------------------------------

    copy = merge(max_radius(1), 0, periodic)
    allocate(extr_map(1 - copy:nlon + copy, nlat))
    copy = merge(1, 0, periodic)
    call local_extrema(extr_map(1 - copy:nlon + copy, :), ind_extr, &
         innermost_level, cyclone, ssh(1 - copy:nlon + copy, :), periodic, &
         my_lbound = [1 - copy, 1])
    ! Assertion: \forall k, ind_extr(1, k) \in \{2 - copy, \dots,
    ! nlon + copy - 1\} and ind_extr(2, k) \in \{2, \dots, nlat - 1\}

    ! Assertion: \forall k, ind_extr(1, k) \in \{1, \dots, nlon\} and
    ! ind_extr(2, k) \in \{2, \dots, nlat - 1\}

    if (periodic) then
       ! Extend in longitude:
       extr_map(1 - max_radius(1):- 1, :) &
            = extr_map(nlon + 1 - max_radius(1):nlon - 1, :)
       extr_map(nlon + 2:nlon + max_radius(1), :) = extr_map(2:max_radius(1), :)
    end if

    number_extr = size(ind_extr, 2)
    allocate(list(number_extr))
    list%innermost_level = innermost_level

    forall (i = 1:number_extr)
       list(i)%extr%coord = invert_proj(ind_extr(:, i))
       ! (Even when periodic, this is within the original NetCDF grid,
       ! that is the grid without duplicated longitudes.)

       list(i)%extr%coord_proj = ind_extr(:, i)
       list(i)%extr%ssh = ssh(ind_extr(1, i), ind_extr(2, i))
       list(i)%cyclone = cyclone(i)
    end forall

  end subroutine set_all_extr

end module set_all_extr_m
